# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

from avocado.core import exceptions
import time
import xmlrpclib
import sys
import logging
import subprocess

EDID_DISABLED = -1
EDID_DEFAULT = 0

_retry_classes = { "I2cBusError" : 3, "InputFlowError" : 3 }

class HID():
    def __init__(self, chameleon):
        self._ch = chameleon
        self._hidport = None

    def _connect(self):
        subprocess.call(['ssh', '-o', 'StrictHostKeyChecking=no', '-o', 'UserKnownHostsFile=/tmp/.fedora-laptop-testing-hostkeys', 'root@%s' % self._ch._host, 'modprobe', 'g_hid'])
        self._kbdport = list(self._ch.iter_ports({'USBKeyboard'}))[0]

    def send_keys(self, keys, enter=False):
        self._ch._p_call("SendHIDEvent", self._kbdport, "keys", keys, enter)


class Chameleon():

    def __init__(self, host, port=9992):
        self._host = host
        self._address = 'http://%s:%s' % (host, port)
        self._p = None
        self._ports = dict()

        self.hid = HID(self)

    def _p_call(self, func, *args, **kwargs):
        if 'delay' in kwargs and kwargs['delay']:
            return getattr(self._p, "Delayed")(func, kwargs['delay'], *args)

        tries = 0
        while True:
            tries += 1
            try:
                return getattr(self._p, func)(*args)
            except xmlrpclib.Fault, f:
                for c in _retry_classes:
                    if c in f.faultString:
                        if tries < _retry_classes[c]:
                            break
                        else:
                            logging.error("Could not handle error while calling %s after %i tries" % (func, tries))
                            raise f
                else:
                    logging.error("Could not handle error while calling %s (%s)" % (func, str(args)))
                    raise f
            except OSError, f:
                # Try two times
                if tries > 1:
                    raise f

    def connect(self):
        if self._p is not None:
            return
        self._p = xmlrpclib.ServerProxy(self._address)

        ports = list(self._p.ProbePorts())
        for port in ports:
            self._ports[port] = self._p.GetConnectorType(port)

        self.hid._connect()


    def physically_plugged(self, port):
        return self._p_call("IsPhysicalPlugged", port)

    def simulate_slow_edid_plug(self, port, final_edid=EDID_DEFAULT, delayms=100):
        self._p_call("ApplyEdid", port, EDID_DISABLED)
        self._p_call("Plug", port)
        time.sleep(delayms / 1000)
        self._p_call("ApplyEdid", port, final_edid)

    def unplug(self, port, delay=None):
        self._p_call("Unplug", port, delay=delay)

    def unplug_all(self):
        for p in self.iter_displays():
            self._p_call("Unplug", p)

    def ensure_unplugged(self, port):
        if self._p_call("IsPlugged", port):
            self._p_call("Unplug", port)
            time.sleep(2)

    def plug(self, port, delay=None):
        self._p_call("Plug", port, delay=delay)

    def set_default_edid(self, port):
        self._p_call("ApplyEdid", port, EDID_DEFAULT)

    def set_edid(self, port, edid):
        # Program the edid and throw it away, we won't need it again
        if edid is not None:
            edid_binary = xmlrpclib.Binary(edid.data)
            edid_id = self._p_call("CreateEdid", edid_binary)
            self._p_call("ApplyEdid", port, edid_id)
            self._p_call("DestroyEdid", edid_id)
        else:
            self._p_call("ApplyEdid", port, EDID_DISABLED)

    def get_resolution(self, port):
        return tuple(self._p_call('DetectResolution', port))

    def screenshot(self, port, x=None, y=None, width=None, height=None):
        return self._p_call("DumpPixels", port, x, y, width, height)

    def get_port_type(self, port):
        return self._ports[port]

    def iter_displays(self):
        for p in self.iter_ports(types={'DP', 'HDMI', 'VGA'}):
            yield p

    def iter_ports(self, types=None):
        for p, t in self._ports.iteritems():
            if types and t not in types:
                continue

            if not self.physically_plugged(p):
                continue

            yield p

    def disconnect(self):
        del self._p
        self._p = None

    def __del__(self):
        self.disconnect()





