# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

# This ensures we use the memory backend for DConf
from . import ui

import re
import sys
import os
from avocado import Test
from avocado.core import exceptions
from . import session
from . import chameleon
from .ui.structs import StateSignal

def batch(func):
    # Note: This duplicates checks from the avocado test class
    def test(self):
        exc = None
        try:
            res = func(self)
        except exceptions.TestBaseException as detail:
            self._func_status = detail.status
            raise
        except AssertionError:
            self._func_status = 'FAIL'
            raise
        except:
            self._func_status = 'ERROR'
            raise
        else:
            self._func_status = 'PASS'
        return res

    return test

def interactive(func):
    """Confirm state from the user still allowing them to mark it as a failure.
    """
    func = batch(func)
    def test(self):
        try:
            res = func(self)
        finally:
            try:
                self.app_ctrl.read_pending()

                if not self.skip_user_confirm:
                    self.update_ui_state()
                    # Unselect test state and select "finish" if available.
                    self.app_ctrl.status.mark_action('finish')
                    self.app_ctrl.status.wait_finished()
            except exceptions.TestBaseException:
                # Re-raise a test cancel/fail exception from the user overriding the
                # actual result (in particular a FAIL result with CANCEL).
                raise
            except:
                details = sys.exc_info()[1]
                self.log.error('Failed to interact with application for user confirmation ({:s})!'.format(details))
                pass
        return res

    return test

class SessionTest(Test):
    """
    A base class for tests which require a GNOME session to run. A UI and some
    event receiver interfaces is spawend automatically in the created session
    and can be controlled.
    The test will fail if the GNOME session or UI fails.

    To use this class you need to set the avocado enable flag in the test.

    Any test function should use the @wrap decorator.
    """

    #: Can be overriden to extend the UI
    ui_class = None
    #: Set if the confirmation should be skipped for some reason
    skip_user_confirm = False

    def setUp(self):
        self.session = session.Session(self)
        self.session.start()
        self.app_ctrl = self.session.app_ctrl
        self.__warned = None
        self.__in_teardown = False
        self._func_status = None

        # Override warn func to be able to show the user that a warning happened
        orig_warn = self.log.warn
        def warn_func(*args, **kwargs):
            orig_warn(*args, **kwargs)

            self.__warned = 'WARN'
            self.update_ui_state()

        self.log.warn = self.log.warning = warn_func

    def tearDown(self):
        self.__in_teardown = True

        try:
            notes = self.app_ctrl.status.get_notes()
            if notes is not None:
                notes = notes.strip()
            if notes:
                self.log.info('The user provided the following notes:\n' + '*'*40 + '\n' + notes + '\n' + '*'*40)
            else:
                self.log.info('The user provided no notes.')
        except:
            details = sys.exc_info()[1]
            self.log.error('Failed to get user notes ({:s})!'.format(details))
        self.session.stop()

    def report_state(self):
        """
        Send the current test state to the test runner process. Overriden to
        force an update of the UI process from here.
        """

        super(SessionTest, self).report_state()
        self.update_ui_state()

    def set_ui_status(self, status):
        self._func_status = status
        self.update_ui_state()

    def update_ui_state(self):
        # Replace status with prelininary_status if not set
        state = self.get_state()
        if state['status'] is None:
            state['status'] = self.preliminary_status
        state['done'] = self.__in_teardown

        self.app_ctrl.send(StateSignal(state))

    def progress(self, ratio):
        self.app_ctrl.status.progress(ratio)

    @property
    def preliminary_status(self):
        status = None
        if self._func_status:
            status = self._func_status

        # Special case if we didn't get a proper status
        if status is None and self.__in_teardown:
            return 'UNKNOWN'

        if status in [None, 'PASS'] and self.__warned:
            status = 'WARN'

        return status

    def recv(self, event):
        pass

class ChameleonTest(SessionTest):
    """
    A base class for tests which require a GNOME session and Chameleon board
    to run.

    To use this class you need to set the avocado enable flag in the test.
    """

    _allow_manual = False

    def setUp(self):
        self.manual = False

        if not 'CHAMELEON_IP' in os.environ:
            if self._allow_manual:
                self.manual = True
                self.info('Chameleon IP was not provided, running manually!')
                return

            self.skip('Chameleon IP was not provided, skipping!')

        try:
            self._chameleon_ip = os.environ['CHAMELEON_IP']
            self.chameleon = chameleon.Chameleon(self._chameleon_ip)
            self.chameleon.connect()

            # Make sure everything is unplugged
            self.chameleon.unplug_all()
        except:
            if self._allow_manual:
                self.manual = True
                self.info('Could not connect to chameleon board, running manually!')
                import traceback
                traceback.print_exc()
                return

            import traceback
            traceback.print_exc()

            self.skip('Could not connect to chameleon board, skipping test!')

        # Setup the session last as a failure would otherwise mean we are not
        # cleaning up.
        SessionTest.setUp(self)
        try:
            self.app_ctrl.monitor.wait_no_external()
        except:
            SessionTest.tearDown(self)


    def tearDown(self):
        SessionTest.tearDown(self)

        if not self.manual:
            # Make sure everything is unplugged
            self.chameleon.unplug_all()
            self.app_ctrl.monitor.wait_no_external()
        else:
            raise NotImplemented("Should prompt for user to unplug?")


