# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import sys
import multiprocessing
import Queue
import subprocess
import sys
import pwd, os
import time
import select

from .structs import *
from gi.repository import Gtk, GLib, Gio

from .status import StatusView

class App(Gtk.Application):

    def __init__(self, test, pipe, start_hidden):
        self._pipe = pipe
        self._start_hidden = start_hidden
        self.test = test

        Gtk.Application.__init__(self,
            application_id="org.fedoraproject.laptop-testing",
            flags=Gio.ApplicationFlags.NON_UNIQUE)

        self.connect("activate", self._activate);

        GLib.unix_fd_add_full(0, self._pipe.fileno(),
                              GLib.IO_IN, self.__ui_recv)

    def _activate(self, app):
        self.main_win = Gtk.Window()
        self.add_window(self.main_win)

        res = self.inhibit(self.main_win,
                           Gtk.ApplicationInhibitFlags.LOGOUT |
                           Gtk.ApplicationInhibitFlags.SWITCH |
                           Gtk.ApplicationInhibitFlags.SUSPEND |
                           Gtk.ApplicationInhibitFlags.IDLE,
                           "Prevent automatic locks from interfering with tests.")

        self._ui = []

        for cls in UIRegister._register:
            obj = cls(self)
            self._ui.append(obj)
            setattr(self, cls.ui_name, obj)

        if self.test.ui_class is not None:
            self.test_ui = self.test.ui_class(self)
            self._ui.append(self.test_ui)
        else:
            self.test_ui = None

        self.status = StatusView(self)
        self.status.set_test(self.test)
        self._ui.append(self.status)

        self.main_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.main_box.set_homogeneous(True)
        self.main_win.add(self.main_box)

        self.main_box.add(self.status)

        test_bin = Gtk.Alignment(xalign=0, yalign=0)
        self.main_box.add(test_bin)

        if self.test_ui is not None and hasattr(self.test_ui, 'get_widget'):
            widget = self.test_ui.get_widget()
            if widget is not None:
                test_bin.add(widget)

        self.main_box.show_all()

        if not self._start_hidden:
            self.main_win.maximize()
            self.main_win.present()
            self.main_win.show()
            self.main_win.connect_after('map-event', lambda *args: self.send(ReadyEvent()))
        else:
            self.send(ReadyEvent())


    def __ui_recv(self, fd, condition):
        try:
            event = self._pipe.recv()
        except EOFError:
            self.quit()
            return False

        if isinstance(event, QuitSignal):
            self.quit()
            return True

        # Pass to all the widgets
        for ui in self._ui:
            if not hasattr(ui, 'recv'):
                pass

            ui.recv(event)

        return True

    def send(self, signal):
        self._pipe.send(signal)

    def fail(self, msg):
        self.send(FailEvent(msg))

    def info(self, msg):
        self.send(InfoEvent(msg))

    def warning(self, msg):
        self.send(WarningEvent(msg))

    def debug(self, msg):
        self.send(DebugEvent(msg))


class AppCtrl(object):

    def __init__(self, test, session, hidden):
        self.test = test
        self.session = session
        self._started = False
        self._start_hidden = hidden
        self._ui_ctrl = []

        for cls in UICtrlRegister._register:
            obj = cls(self)
            self._ui_ctrl.append(obj)
            setattr(self, cls.ui_name, obj)


        self._pipe, pipe_app = multiprocessing.Pipe(True)

        def wrapper_func():
            # Wrap the main UI, this function must never return!

            def except_hook(exctype, value, tb):
                import traceback
                info = ''.join(traceback.format_exception(exctype, value, tb, limit=10))

                # Unfortunately sometimes the data seems not to be read back
                # properly. So write out to stderr unconditionally
                sys.stderr.write('Got unhandled exception in user application!\n')
                sys.stderr.write('Exception was:\n')
                sys.stderr.write(info)
                sys.stderr.flush()
                try:
                    pipe_app.send(Exception(info))
                except Exception, e:
                    pass

                # Note: no cleanup to prevent avocado Test structures to be
                #       cleaned.
                os._exit(1)

            sys.excepthook = except_hook
            res = 1

            try:
                # Close the remote end of the pipe
                self._pipe.close()

                # And create and start the app
                app = App(test, pipe_app, self._start_hidden)
                res = app.run()

            except:
                import traceback
                exc_type, exc_value, exc_traceback = sys.exc_info()

                except_hook(exc_type, exc_value, exc_traceback)

            finally:
                # Note: no cleanup to prevent avocado Test structures to be
                #       cleaned.
                os._exit(res)

        self._proc = self.session.spawn_process(wrapper_func)
        pipe_app.close()

    def send(self, signal):
        self._pipe.send(signal)

    def stop(self):
        if self._proc is None:
            return

        try:
            self.send(QuitSignal())
        except (IOError, UIDoneError) as e:
            pass

        try:
            self.read_pending()
            self._proc.join(2)
            self.read_pending()

            # SIGTERM doesn't seem to be delivered?
            if self._proc.is_alive():
                os.kill(self._proc.pid, 9)
        except (IOError, UIDoneError) as e:
            pass
        self._proc = None

    def _recv(self):
        try:
            event = self._pipe.recv()
        except EOFError:
            raise UIDoneError("The user session application failed!")

        if isinstance(event, Exception):
            raise UIError("Exception in session application! Original exception follows\n" + str(event))

        if isinstance(event, WarningEvent):
            self.test.log.warning(event.warning)

        if isinstance(event, DebugEvent):
            self.test.log.debug(event.debug)

        if isinstance(event, InfoEvent):
            self.test.log.info(event.info)

        if isinstance(event, FailEvent):
            if event.no_user_confirm:
                self.test.skip_user_confirm = True
            self.test.fail(event.fail)

        if isinstance(event, CancelEvent):
            if event.no_user_confirm:
                self.test.skip_user_confirm = True
            self.test.cancel(event.cancel)

        if isinstance(event, ReadyEvent):
            self._started = True

        self.test.recv(event)

        for ui_ctrl in self._ui_ctrl:
            if not hasattr(ui_ctrl, 'recv'):
                pass

            ui_ctrl.recv(event)

        return event

    def wait_event(self, timeout=0, filt=None):
        start = time.time()
        end = start + timeout

        while True:
            try:
                if timeout > 0:
                    res = self._pipe.poll(timeout)
                else:
                    res = self._pipe.poll(5)
            except IOError:
                raise UIDoneError("The user session application failed!")

            if res:
                event = self._recv()
                if filt is not None:
                    if not filt(event):
                        continue

                return event

            if timeout > 0:
                timeout = end - time.time()
                if timeout <= 0:
                    raise TimeoutError()
            continue

        raise TimeoutError()

    def wait_started(self, timeout=15):
        if self._started:
            return True

        self.wait_event(timeout, filt=lambda event: self._started)

    def read_pending(self):
        while self._pipe.poll():
            event = self._recv()

    def select(self, r=[], w=[], x=[], timeout=None):
        fd = self.fileno()
        r, w, x = select.select(r + [fd], [], [], timeout)

        self.read_pending()
        if fd in r:
            r.remove(fd)
        return r, w, x

    def get_event(self):
        return self._recv()

    def fileno(self):
        return self._pipe.fileno()


