# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

from gi.repository import Gtk, Gdk, GLib
from .structs import *

class Monitor(object):
    def __init__(self):
        pass

    @classmethod
    def from_monitor(cls, mon):
        res = cls()

        geometry = mon.get_geometry()
        res.height_mm = mon.get_height_mm()
        res.width_mm = mon.get_width_mm()
        res.manufacturer = mon.get_manufacturer()
        res.model = mon.get_model()
        res.scale = mon.get_scale_factor()
        res.app_x = geometry.x
        res.app_y = geometry.y
        res.app_width = geometry.width
        res.app_height = geometry.height
        # XXX: This is probably still wrong for 1.5 factor scaling!
        res.width = geometry.width * res.scale
        res.height = geometry.height * res.scale

        res.hash = hash(mon)

        return res

    @classmethod
    def from_screen(cls, screen, mon):
        res = cls()

        geometry = screen.get_monitor_geometry(mon)
        res.height_mm = screen.get_monitor_height_mm(mon)
        res.width_mm = screen.get_monitor_width_mm(mon)
        res.manufacturer = 'UNKNOWN'
        res.model = 'Plug: ' + screen.get_monitor_plug_name(mon)
        res.scale = screen.get_monitor_scale_factor(mon)
        res.app_x = geometry.x
        res.app_y = geometry.y
        res.app_width = geometry.width
        res.app_height = geometry.height
        # XXX: This is probably still wrong for 1.5 factor scaling!
        res.width = geometry.width * res.scale
        res.height = geometry.height * res.scale

        # Use the plug as a unique identifier
        res.hash = screen.get_monitor_plug_name(mon)

        return res

    def __repr__(self):
         return 'Monitor(%s, %s, %i, %i, scale=%i)' % (self.manufacturer, self.model, self.app_width, self.app_height, self.scale)


class MonitorEvent(Event):
    def __init__(self, event, monitor):
        self.event = event
        self.monitor = monitor


class MonitorWatcher(TestUI):

    __metaclass__ = UIRegister
    ui_name = "monitor"

    def __init__(self, app):
        self.app = app
        self._attached = {}

        # Send a reset of the test side monitor list
        self.app.send(MonitorEvent('reset', None))

        display = Gdk.Display.get_default()
        try:
            display.connect('monitor-added', self.monitor_added_cb)
            display.connect('monitor-removed', self.monitor_removed_cb)

            # Simulate adding monitors so that the deamon knows about all of them
            n = display.get_n_monitors()
            for i in xrange(n):
                mon = display.get_monitor(i)
                self.monitor_added_cb(display, mon)
        except TypeError:
            self.app.test.log.info('Using old API to query monitors!')

            screen = Gdk.Screen.get_default()
            screen.connect('monitors-changed', self.monitors_changed_cb)
            self.monitors_changed_cb(screen)

    def send_monitor_event(self, event, mon):
        monitor = Monitor.from_monitor(mon)
        self.app.send(MonitorEvent(event, monitor))

    def monitor_added_cb(self, obj, mon):
        self.send_monitor_event('added', mon)

    def monitor_removed_cb(self, obj, mon):
        self.send_monitor_event('removed', mon)

    # Old GTK+ API
    def monitors_changed_cb(self, screen):
        n = screen.get_n_monitors()

        current = {}
        for mon in xrange(screen.get_n_monitors()):
            monitor = Monitor.from_screen(screen, mon)
            current[monitor.hash] = monitor

            self.app.send(MonitorEvent('added', monitor))

        for output, monitor in self._attached.iteritems():
            if not output in current:
                self.app.send(MonitorEvent('removed', monitor))

        self._attached = current

class MonitorWatcherCtrl(TestUICtrl):

    __metaclass__ = UICtrlRegister
    ui_name = "monitor"

    def __init__(self, app_ctrl):
        self.app_ctrl = app_ctrl
        self.test = app_ctrl.test
        self.log = self.test.log
        self._remote_monitors = {}

    def recv(self, event):
        if not isinstance(event, MonitorEvent):
            return

        if event.event == 'reset':
            self._remote_monitors = {}
        elif event.event == 'added':
            mon = event.monitor
            self._remote_monitors[mon.hash] = mon
            self.log.debug("Monitor added event %s", str(mon))

        elif event.event == 'removed':
            # Retrieve cached information
            mon = self._remote_monitors[event.monitor.hash]
            del self._remote_monitors[event.monitor.hash]
            self.log.debug("Monitor removed event %s", str(mon))

        else:
            raise AssertionError("Do not understand MonitorEvent of type %s" % event.event)

    def _match_monitor(mon, filt):
        for k, v in kwargs:
            if getattr(mon, k) != v:
                return False
        return True

    def exists(self, **kwargs):
        for mon in self._remote_monitors.iter_values():
            if _monitor_matches(mon, kwargs):
                return mon
        else:
            # No match found, so return.
            return False
        return False

    def wait_missing(self, timeout=10, **kwargs):
        # Check whether the monitor exists, note that we just re-check them
        # all here for simplicity (the event monitor has been called already)
        self.app_ctrl.wait_event(timeout, lambda event: not self.exists(**kwargs))

    def wait_exists(self, timeout=10, **kwargs):
        # Check whether the monitor exists, note that we just re-check them
        # all here for simplicity (the event monitor has been called already)
        self.app_ctrl.wait_event(timeout, lambda event: self.exists(**kwargs))
        return self.exists(**kwargs)

    def wait_added(self, timeout=10):
        # Check whether the monitor exists, note that we just re-check them
        # all here for simplicity (the event monitor has been called already)
        event = self.app_ctrl.wait_event(timeout, lambda event: isinstance(event, MonitorEvent) and event.event == 'added')
        return event.monitor

    def wait_removed(self, timeout=10):
        # Check whether the monitor exists, note that we just re-check them
        # all here for simplicity (the event monitor has been called already)
        event = self.app_ctrl.wait_event(timeout, lambda event: isinstance(event, MonitorEvent) and event.event == 'removed')
        return event.monitor

    def wait_no_external(self, timeout=10):
        self.wait_count(1)

    def wait_count(self, count, timeout=10):
        # Check whether the monitor exists, note that we just re-check them
        # all here for simplicity (the event monitor has been called already)
        if len(self._remote_monitors) == count:
            return
        self.app_ctrl.wait_event(timeout, lambda event: len(self._remote_monitors) == count)

