#!/usr/bin/env python
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import os
import os.path
import gi
import json
import re

gi.require_version('Gtk', '3.0')
try:
    gi.require_version('WebKit2', '4.0')
except ValueError:
    gi.require_version('WebKit2', '3.0')

import docutils.core

from gi.repository import Gtk, GLib, Gdk
from gi.repository import WebKit2

from .structs import Signal, StateSignal, Event, TestUICtrl, UICtrlRegister, FailEvent, WarningEvent, CancelEvent
from .. import utils
from ..utils.ui import status_to_icon, status_to_description

class MarkActionSignal(Signal):
    def __init__(self, action):
        self.action = action

class ProgressSignal(Signal):
    def __init__(self, progress, message=None):
        self.progress = progress
        self.message = message

class AllowConfirmSignal(Signal):
    def __init__(self, allow, cont=False):
        self.allow = allow
        self.cont = cont

class NoteUpdatedEvent(Event):
    def __init__(self, note):
        self.note = note

class ConfirmEvent(Event):
    def __init__(self, cont=False):
        self.cont = cont

class StatusView(Gtk.Box):

    def __init__(self, app):
        super(StatusView, self).__init__(orientation=Gtk.Orientation.VERTICAL)

        self.app = app

        uifile = os.path.join(utils.get_data_dir(), 'ui', 'status.ui')
        builder = Gtk.Builder.new_from_file(uifile)
        builder.connect_signals(self)
        status_view = builder.get_object('status-view')
        status_view.show_all()

        self.test_title = builder.get_object('title')
        self.confirm_btn = builder.get_object('confirm')
        self.continue_btn = builder.get_object('continue')
        self._confirm_allowed = False
        self.fail_btn = builder.get_object('fail')
        self.notes_textview = builder.get_object('notes')
        self.notes = self.notes_textview.get_buffer()
        self.notes.connect('changed', self.notes_changed_cb)

        # We cause the default action if the user presses Shift+Enter
        self.notes_textview.connect('key-press-event', self.textview_key_press_cb)

        self.status_img = builder.get_object('status-image')
        self.status_label = builder.get_object('status-label')


        self.pack_start(status_view, expand=False, fill=True, padding=0)

        view = self._create_webview()
        self.pack_start(view, fill=True, expand=True, padding=0)
        self._current_action = None
        self._progress = (None, None)

    def notes_changed_cb(self, buf):
        start = self.notes.get_start_iter()
        end = self.notes.get_end_iter()
        self.app.send(NoteUpdatedEvent(self.notes.get_text(start, end, False)))

    def textview_key_press_cb(self, textview, event):
        if event.keyval == Gdk.KEY_Return and (event.state & Gdk.ModifierType.SHIFT_MASK):
            textview.get_toplevel().activate_default()
            return True

    def confirm_clicked_cb(self, btn):
        self.app.send(ConfirmEvent())

    def continue_clicked_cb(self, btn):
        self.app.send(ConfirmEvent(cont=True))
        self.continue_btn.set_sensitive(False)

    def cancel_clicked_cb(self, btn):
        self.app.send(CancelEvent('User cancelled the test!', no_user_confirm=True))

    def warn_clicked_cb(self, btn):
        self.app.send(WarningEvent('User send a warning!'))

    def fail_clicked_cb(self, btn):
        self.app.send(FailEvent('User marked the test as a failure!', no_user_confirm=True))

    def set_test(self, test):
        # Create a default title from the test name
        self.title = re.sub('^.*:', '', test.name.name)
        docstring = test.__doc__

        self._current_action = None
        self._progress = (None, None)

        resource_path = os.path.join(utils.get_data_dir(), 'html')

        self._load_docstring(docstring, resource_path)

        self.view.load_html(self.html, 'file://' + resource_path + '/')
        self.test_title.set_label(self.title)

    def mark_action(self, action=None):
        self._current_action = action
        if action:
            self.view.run_javascript('mark_action("%s")' % action)
        else:
            self.view.run_javascript('mark_action()')

    def set_progress(self, progress, message=None):
        self._progress = (progress, message)
        if progress is None:
            return

        if not message:
            self.view.run_javascript('set_progress(%f)' % progress)
        else:
            self.view.run_javascript('set_progress(%f, %s)' % (progress, json.dumps(message)))

    def recv(self, event):
        if isinstance(event, MarkActionSignal):
            self.mark_action(event.action)

            return True

        elif isinstance(event, ProgressSignal):
            self.set_progress(event.progress, event.message)

        elif isinstance(event, AllowConfirmSignal):
            if event.cont:
                btn = self.continue_btn
            else:
                self._confirm_allowed = event.allow
                btn = self.confirm_btn

            btn.set_sensitive(event.allow)
            self.notes_textview.grab_focus()
            btn.grab_default()

        elif isinstance(event, StateSignal):
            status = event.state['status']
            # Consider passed if status is alright and user is allowed to confirm
            if status is None and self._confirm_allowed:
                status = 'PASS'

            icon = status_to_icon(status)
            status_string = status_to_description(status, True)

            self.status_img.props.gicon = icon
            self.status_label.set_text(status_string)

            self.confirm_btn.set_sensitive(event.state['done'] or self._confirm_allowed)
            self.fail_btn.set_sensitive(not event.state['done'])

    def _load_event_cb(self, view, event):
        if event != WebKit2.LoadEvent.FINISHED:
            return

        self.mark_action(self._current_action)
        self.set_progress(*self._progress)

    def _load_docstring(self, docstring, resource_path, encoding='utf-8'):
        parts = docutils.core.publish_parts(docstring, writer_name='html')

        template = open(os.path.join(resource_path, 'testcase.template')).read()

        title_match = re.search(r'.*<title>(?P<title>.*)</title>', parts['head'])
        if title_match is not None and title_match.group('title'):
            self.title = title_match.group('title')

        parts['html_head'] = parts['html_head'] % encoding

        self.html = template % parts

    def _create_webview(self):
        self.view = view = WebKit2.WebView()
        view.set_size_request(-1, 200)

        view.connect('load-changed', self._load_event_cb)

        settings = view.get_settings()
        # Enable devloper tools
        settings.props.enable_developer_extras = True
        settings.props.enable_fullscreen = False
        # Required to load jquery locally
        try:
            settings.props.allow_file_access_from_file_urls = True
            settings.props.allow_universal_access_from_file_urls = True
        except AttributeError:
            pass

        return view


class StatusViewCtrl(TestUICtrl):

    __metaclass__ = UICtrlRegister
    ui_name = "status"

    def __init__(self, app_ctrl):
        TestUICtrl.__init__(self, app_ctrl)
        self.__continue = False
        self.__finished = False
        self.__user_note = ''

    def mark_action(self, action=None):
        self.app_ctrl.send(MarkActionSignal(action))

    def progress(self, progress, message=None):
        self.app_ctrl.send(ProgressSignal(progress, message))

    def recv(self, event):
        if isinstance(event, ConfirmEvent):
            # The poll in tearDown will return immediately if this happens
            # during the test.
            if event.cont:
                self.__continue = True
            else:
                self.__finished = True

        elif isinstance(event, NoteUpdatedEvent):
            self.__user_note = event.note

    def allow_finish(self, allow=True):
        self.app_ctrl.send(AllowConfirmSignal(allow))
        self.app_ctrl.test.update_ui_state()

    def allow_continue(self, allow=True):
        self.__continue = False
        self.app_ctrl.send(AllowConfirmSignal(allow, cont=True))
        self.app_ctrl.test.update_ui_state()

    @property
    def finished(self):
        return self.__finished

    @property
    def continued(self):
        return self.__continue

    def wait_finished(self):
        self.allow_finish()
        while not self.__finished:
            self.app_ctrl.wait_event()

    def wait_continue(self):
        self.__continue = False
        self.app_ctrl.send(AllowConfirmSignal(True, cont=True))
        while not self.__continue:
            self.app_ctrl.wait_event()
        self.__continue = False

    def get_notes(self):
        # Flush out everything, just in case
        self.app_ctrl.read_pending()
        return self.__user_note

