#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>


import struct
import sys
import os
import os.path
from collections import defaultdict

class Vendor():
    def __init__(self, vid, name):
        self.vid = vid
        self.name = name.strip()
        self.products = {}

class Product():
    def __init__(self, pid, name):
        self.pid = pid
        self.name = name.strip()

vendors = dict()

usb_ids_file = "/usr/share/hwdata/usb.ids"
def parse_usb_ids():
    cur_vendor = None
    for line in open(usb_ids_file, 'r').xreadlines():
        if line.startswith('#'):
            continue

        if line.startswith('\t\t'):
            # We don't care about endpoints here
            pass
        elif line.startswith('\t'):
            if cur_vendor is None:
                continue

            line = line.strip()
            try:
                l = line.split('  ', 1)
                assert len(l[0]) == 4
                pid = int(l[0], 16)
            except:
                # Not a valid hex digit
                cur_vendor = None
                continue

            p = Product(pid, l[1])
            cur_vendor.products[pid] = p
        else:
            l = line.strip()
            try:
                l = line.split('  ', 1)
                assert len(l[0]) == 4
                vid = int(l[0], 16)
            except:
                # Not a valid hex digit
                cur_vendor = None
                continue

            cur_vendor = Vendor(vid, l[1])
            vendors[vid] = cur_vendor

parse_usb_ids()

def find_usb_vendor(vid):
    print(vid)
    try:
        return vendors[vid].name
    except KeyError:
        return None

def find_usb_prod(vid, pid):
    try:
        return vendors[vid].products[pid].name
    except KeyError:
        return None

class USBSlot(object):

    """Represent a USB slot of any version and allows querying information about
    the device that is attached to it. If no device is present, responds with
    the known capabilities instead."""

    def __init__(self, parent, port, busses):
        self._port = port
        self._parent = parent
        self._slot_busses = busses
        self._handles = []

        self.slot_version_str
        self.version_str

    @property
    def parent(self):
        return self._parent

    @property
    def port(self):
        return self._port

    @property
    def slot_version(self):
        return int(float(self.version_str) * 100)

    @property
    def slot_version_str(self):
        try:
            return self._slot_version_str
        except:
            pass

        version = " 1.00"
        for ver in self.parent.sysfsdata('version', busses=self._slot_busses):
            if ver > version:
                version = ver

        self._slot_version_str = version.strip()
        return self._slot_version_str

    version = slot_version
    version_str = slot_version_str

    def paths(self, parent=False, descr=None, busses=None):
        if busses is None:
            busses = self.busses

        for path in self._parent.paths(parent=True, busses=busses):
            if descr:
                yield path + '{:d}:{:s}'.format(self._port, descr)
            elif parent:
                yield path + '{:d}.'.format(self._port)
            else:
                yield path + '{:d}'.format(self._port)

    def sysfs(self, descr=None, busses=None, f=None):
        for path in self.paths(descr=descr, busses=busses):
            if f:
                yield os.path.join('/sys/bus/usb/devices', path, f)
            else:
                yield os.path.join('/sys/bus/usb/devices', path)

    def sysfsdata(self, f, busses=None):
        for path in self.paths(busses=busses):
            yield open(os.path.join('/sys/bus/usb/devices', path, f)).read().strip('\n')

    def has_device(self):
        have_device = False
        for path in self.sysfs():
            if os.path.exists(path):
                have_device = True

        return have_device

    def get_hub_port_info(self):
        assert self.is_hub()

    @property
    def info_str(self):
        return str(self)

    def __repr__(self):
        return 'Port %i (USB %s)' % (self.port, self.slot_version_str)

    def show_tree(self, indent=''):
        sys.stdout.write(indent + str(self) + '\n')

    @property
    def busses(self):
        return self._slot_busses

    def __eq__(self, other):
        if self.parent != other.parent:
            return False
        if self._port != other._port:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

class USBDevice(USBSlot):
    def __init__(self, *args):
        USBSlot.__init__(self, *args)
        assert self.has_device()

        self.idProduct = int(list(self.sysfsdata('idProduct'))[-1], 16)
        self.idVendor = int(list(self.sysfsdata('idVendor'))[-1], 16)

        try:
            self.product = list(self.sysfsdata('product'))[-1]
        except IOError:
            self.product = find_usb_prod(self.idVendor, self.idProduct)
            if not self.product:
                self.product = 'Product %04X' % self.idProduct
        try:
            self.manufacturer = list(self.sysfsdata('manufacturer'))[-1]
        except IOError:
            self.manufacturer = find_usb_vendor(self.idVendor)
            if not self.manufacturer:
                self.manufacturer = 'Vendor %04X' % self.idProduct

        self.busses

    def is_hub(self):
        for devicecls in self.sysfsdata('bDeviceClass'):
            if devicecls != '09':
                return False

        return True

    @property
    def busses(self):
        try:
            return self._busses
        except:
            pass

        busses = list()
        for bus in self._slot_busses:
            for path in self.paths(busses=[bus]):
                if os.path.exists(os.path.join('/sys/bus/usb/devices', path)):
                    busses.append(bus)
        self._busses = busses
        return self._busses

    @property
    def info_str(self):
        if self.product is None and self.manufacturer is None:
            info = find_usb_prod(self.idVendor, self.idProduct).strip()
        elif self.manufacturer is None:
            info = find_usb_vendor(self.idVendor).strip() + ' ' + self.product.strip()
        else:
            info = self.manufacturer.strip() + ' ' + self.product.strip()

        return info

    @property
    def version(self):
        return int(float(self.version_str) * 100)

    @property
    def version_str(self):
        try:
            return self._version_str
        except:
            pass

        version = " 1.00"
        for ver in self.sysfsdata('version'):
            if ver > version:
                version = ver

        self._version_str = version.strip()
        return self._version_str

    def __repr__(self):
        return 'Port %i (USB %s, Device USB %s): %s' % (self.port, self.slot_version_str, self.version_str, self.info_str)


class USBHub(USBDevice):
    def __init__(self, *args):
        USBDevice.__init__(self, *args)
        assert self.is_hub()
        self._ports = []

        self.enumerate_slots()

    def enumerate_slots(self):
        ports = 0
        on_bus = dict()

        for bus in self.busses:
            path = list(self.sysfs(descr='1.0', busses=[bus]))[0]
            prefix = list(self.paths(parent=False, busses=[bus]))[0]

            port_count = 0
            prefix += '-port'.format(bus)
            prefix_len = len(prefix)
            for p in os.listdir(path):
                if not p.startswith(prefix):
                    assert '-port' not in p
                    continue

                p = int(p[prefix_len:])
                port_count = max(port_count, p)

            ports = max(port_count, ports)
            on_bus[bus] = port_count

        for p in xrange(1, ports+1):
            busses = []
            for b, ports in on_bus.iteritems():
                if p <= ports:
                    busses.append(b)

            slot = USBSlot(self, p, busses)
            if slot.has_device():
                try:
                    dev = USBDevice(self, p, busses)

                    if dev.is_hub():
                        dev = USBHub(self, p, busses)
                    slot = dev
                except IOError:
                    pass
                except OSError:
                    pass

            self._ports.append(slot)

    @property
    def ports(self):
        return self._ports

    def show_tree(self, indent=''):
        sys.stdout.write(indent + str(self) + '\n')
        indent += '  '
        for p in self._ports:
            p.show_tree(indent)


class USBRoot(USBHub):
    def __init__(self, busses):
        USBHub.__init__(self, None, None, busses)

    def paths(self, parent=False, descr=None, busses=None):
        if busses is not None:
            assert set(self._slot_busses).issuperset(busses)
        else:
            busses = self.busses

        for bus in busses:
            if descr:
                yield '%i-0:' % bus + descr
            elif parent:
                yield '%i-' % bus
            else:
                yield 'usb%i' % bus

    @property
    def slot_version_str(self):
        return self.version_str

    def libusb_paths(self):
        for bus in self.busses:
            yield (bus, )

    def __eq__(self, other):
        return self.busses == other.busses

    def __repr__(self):
        return 'Root for busses %s (USB %s): %s' % (', '.join(str(b) for b in self.busses), self.version_str, self.info_str)

def get_roots():
    device_busses = defaultdict(lambda : [])
    for bus in os.listdir('/sys/bus/usb/devices'):
        if not bus.startswith('usb'):
            continue

        device = os.path.dirname(os.readlink(os.path.join('/sys/bus/usb/devices', bus)))

        device_busses[device].append(int(bus[3:]))

    for device, busses in sorted(device_busses.iteritems()):
        # The sysfs tree might be disappearing/appearing
        try:
            yield USBRoot(sorted(busses))
        except IOError:
            pass
        except OSError:
            pass

if __name__ == '__main__':
    for root in get_roots():
        root.show_tree()
