#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import os
import os.path
import re
import random

from avocado import main
from avocado.core import exceptions
from fed_laptoptest import chameleon

from fed_laptoptest import ui
from fed_laptoptest.test import ChameleonTest, batch
from fed_laptoptest.edid import Edid

_EDID_TYPES = {'HDMI': {'HDMI', 'MHL', 'DVI'},
               'DP': {'DP'},
               'VGA': {'VGA'}}


class EDIDStress(ChameleonTest):
    """
    .. title:: EDID profile stress test

    This test uses the chameleon board to test a lot of EDID profiles against
    the laptop in question. We only check that the resolution that the
    chameleon board sees matches the resolution in userspace.

    .. class:: action-wait

        Please wait while the test completes.

    :avocado: enable
    :avocado: tags=chameleon,standalone

    :param edids: The EDID test set to use
    :param edid_re: Test EDID names matching expression
    :categories: graphics
    """

    @batch
    def test(self):
        """
        Test plugging in with slow edid a number of times.
        """
        testset = self.params.get('testset', default='daily')
        random_subset = self.params.get('random', default=10)
        edid_re = self.params.get('edid_re', default='.*')

        dirname = os.path.join(self.datadir, 'edids', testset)

        self.app_ctrl.status.mark_action('wait')

        port_edids = {}
        count = 0

        for port in self.chameleon.iter_displays():
            porttype = self.chameleon.get_port_type(port)

            # Find all the relevant EDID files
            type_re = '|'.join(_EDID_TYPES[porttype])#
            regexp = r'%s_(%s)\.txt' % (edid_re, type_re)
            self.log.debug("Matching EDID files against RE: %s" % regexp)
            regexp = re.compile(regexp, flags=re.IGNORECASE)

            edids = []
            for edidfile in os.listdir(dirname):
                if not regexp.match(edidfile):
                    continue
                edids.append(edidfile)

            if random_subset:
                random.shuffle(edids)
                edids = edids[0:10]

            port_edids[port] = edids
            count += len(edids)

        done = 0
        for port, edids in port_edids.iteritems():
            for edidfile in edids:
                self.app_ctrl.status.progress(done / float(count), 'Testing %s (%i of %i)' % (edidfile, done + 1, count))
                self.log.info("Testing EDID %s", edidfile)

                edidpath = os.path.join(dirname, edidfile)

                edid = Edid.from_file(edidpath)
                self.chameleon.set_edid(port, edid)
                self.chameleon.plug(port)
                mon = self.app_ctrl.monitor.wait_added()

                resolution = self.chameleon.get_resolution(port)

                if mon.width != resolution[0] or mon.height != resolution[1]:
                    raise AssertionError('Resolution did not match for EDID %s (%i, %i, %i, %i)'
                                         % (edidfile, mon.width, resolution[0], mon.height, resolution[1]))

                self.chameleon.unplug(port)
                self.app_ctrl.monitor.wait_no_external()
                done += 1

        self.app_ctrl.status.progress(1, 'Test done')

        self.log.info("All EDIDs tested successfully")

if __name__ == "__main__":
    main()

