#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import os
import os.path
import re
import evdev
import select

from avocado import main
from avocado.core import exceptions

from fed_laptoptest.test import SessionTest, interactive
from fed_laptoptest.hwinfo import HWinfo
from fed_laptoptest.ui import Signal, Event, TestUI

from gi.repository import Gtk, GLib, Gdk


modifier_keycodes = {
        evdev.ecodes.KEY_LEFTALT    : 'Alt',
        evdev.ecodes.KEY_LEFTCTRL    : 'Ctrl',
        evdev.ecodes.KEY_LEFTMETA    : 'Meta',
        evdev.ecodes.KEY_LEFTSHIFT   : 'Shift',

        evdev.ecodes.KEY_RIGHTALT    : 'Alt',
        evdev.ecodes.KEY_RIGHTCTRL    : 'Ctrl',
        evdev.ecodes.KEY_RIGHTMETA    : 'Meta',
        evdev.ecodes.KEY_RIGHTSHIFT   : 'Shift',
    }

# The meaning of some keys is a bit cryptic, so give them a saner description
default_descriptions = {
    'F20' : 'Microphone Mute'
}

class Key(object):
    def __init__(self):
        self.scancode = -1
        self.codes = '-'
        self.modifiers = set()
        self.devname = ''
        self.location = ''
        self.description = ''
        self.dbmissing = False
        self.found = False
        self.tested = False

class KeySignal(Signal):
    def __init__(self, key):
        self.key = key

class KeyChangedEvent(Event):
    def __init__(self, scancode, devname, modifiers, **kwargs):
        self.scancode = scancode
        self.devname = devname
        self.modifiers = modifiers
        self.changes = kwargs

class KeyDeleteEvent(Event):
    def __init__(self, scancode, devname, modifiers):
        self.scancode = scancode
        self.devname = devname
        self.modifiers = modifiers

class EVDevKeysUI(TestUI):

    def __init__(self, *args, **kwargs):
        TestUI.__init__(self, *args, **kwargs)
        self.unselect_id = 0

    def recv(self, signal):
        if not isinstance(signal, KeySignal):
            return

        key = signal.key

        data = []
        data.append(key.scancode)
        data.append(','.join(key.codes))
        data.append('+'.join(sorted(key.modifiers)))
        data.append(key.devname)
        data.append(key.location)
        data.append(key.description)

        # Non-standard icon ...
        icon = 'gtk-yes'
        status = 'Tested'
        if key.dbmissing:
            icon = 'dialog-warning'
            status = 'Not in DB'

        if not key.found:
            icon = 'dialog-error'
            status = 'Not yet pressed'

        if key.tested == False:
            icon = 'dialog-question'
            status = 'Not tested'
            if key.found:
                status = 'Not tested (pressed)'

        data.append(icon)
        data.append(status)

        self.treeview.grab_focus()
        it = self.liststore.get_iter_first()
        while it:
            row = self.liststore[it]

            if row[3] != '' and row[3] != key.devname:
                it = self.liststore.iter_next(it)
                continue

            if row[0] == -1 or row[0] != key.scancode:
                it = self.liststore.iter_next(it)
                continue

            if row[2] != data[2]:
                it = self.liststore.iter_next(it)
                continue

            # Do not execute else branch as we updated a row
            break
        else:
            it = self.liststore.append(data)

        selection = self.treeview.get_selection()
        selection.select_iter(it)
        self.liststore.set_row(it, data)
        GLib.idle_add(self.scroll_to_selection)
        self.queue_unselect()

    def queue_unselect(self):
        if self.unselect_id:
            GLib.source_remove(self.unselect_id)
        self.unselect_id = GLib.timeout_add(1000, self.unselect)

    def unselect(self):
        selection = self.treeview.get_selection()
        selection.unselect_all()

        self.unselect_id = 0
        return False

    def scroll_to_selection(self):
        selection = self.treeview.get_selection()
        model, it = selection.get_selected()
        if it is not None:
            path = self.liststore.get_path(it)
            self.treeview.scroll_to_cell(path, None, False, 0, 0)

    def get_iter_info(self, it):
        scancode = self.liststore[it][0]
        if self.liststore[it][2]:
            modifiers = set(self.liststore[it][2].split('+'))
        else:
            modifiers = set()
        devname = self.liststore[it][3]

        return scancode, devname, modifiers

    def location_edited_cb(self, renderer, path, new_text):
        it = self.liststore.get_iter(path)
        info = self.get_iter_info(it)
        self.app.send(KeyChangedEvent(*info, location=new_text))

    def description_edited_cb(self, renderer, path, new_text):
        it = self.liststore.get_iter(path)
        info = self.get_iter_info(it)
        self.app.send(KeyChangedEvent(*info, description=new_text))

    def treeview_key_press_cb(self, treeview, event):
        if event.keyval == Gdk.KEY_Delete:
            selection = treeview.get_selection()
            model, it = selection.get_selected()
            if it is not None:
                info = self.get_iter_info(it)
                self.liststore.remove(it)
                self.app.send(KeyDeleteEvent(*info))

                return True

    def get_widget(self):

        uifile = os.path.join(self.app.test.datadir, 'evdev_keys_list.ui')

        self.builder = Gtk.Builder.new_from_file(uifile)
        self.widget = self.builder.get_object("evdev_keys_ui")
        self.treeview = self.builder.get_object("detected_keys_treeview")
        self.liststore = self.builder.get_object("detected_keys_store")

        renderer = Gtk.CellRendererText()
        col_key = Gtk.TreeViewColumn("Key", renderer, text=1)
        renderer = Gtk.CellRendererText()
        col_mod = Gtk.TreeViewColumn("Modifier", renderer, text=2)
        renderer = Gtk.CellRendererText()
        col_device = Gtk.TreeViewColumn("Device", renderer, text=3)
        renderer = Gtk.CellRendererText(editable=True)
        renderer.connect('edited', self.location_edited_cb)
        col_loc = Gtk.TreeViewColumn("Location", renderer, text=4)
        renderer = Gtk.CellRendererText(editable=True)
        renderer.connect('edited', self.description_edited_cb)
        col_desc = Gtk.TreeViewColumn("Description", renderer, text=5)

        layout = Gtk.CellAreaBox()
        renderer = Gtk.CellRendererPixbuf()
        layout.pack_start(renderer, False, True, True)
        layout.add_attribute(renderer, "icon-name", 6)
        renderer = Gtk.CellRendererText()
        layout.pack_start(renderer, True, True, False)
        layout.add_attribute(renderer, "text", 7)

        renderer_status = Gtk.CellRendererText()

        col_status = Gtk.TreeViewColumn.new_with_area(layout)
        col_status.set_title("Status")

        self.treeview.connect('key-press-event', self.treeview_key_press_cb)

        self.treeview.append_column(col_status)
        self.treeview.append_column(col_key)
        self.treeview.append_column(col_mod)
        self.treeview.append_column(col_device)
        self.treeview.append_column(col_loc)
        self.treeview.append_column(col_desc)
        self.liststore.set_sort_column_id(1, Gtk.SortType.ASCENDING)

        return self.widget


class EVDevKeys(SessionTest):
    """
    .. title:: Special Keys Test

    In this test we check that all key presses are registered correctly. This
    includes almost all special keys including the power button. The test
    prevents any other system component from registering the key presses.

    When available, a list of known buttons for this hardware is displayed.

    .. class:: action-grabbed

        Please press all special keys the laptop has. Do not press special keys
        on external keyboards (e.g. volume up/down). Please press either *Enter*
        or *Esc* to edit the result or finish the test.

        .. note::
            Do *not* press keyboard combinations as these might be send by
            some special keys.

    .. class:: action-query

        Please edit the key information (location, description) for keys. You
        can also delete keys again if those were misdetections. If you have keys
        that are not listed that you expect to be working then please say so in
        the comment field.

        Hit *"Continue"* to start testing again.

    :avocado: enable
    :avocado: tags=manual
    """

    ui_class = EVDevKeysUI

    def setUp(self):
        SessionTest.setUp(self)

        self.hwinfo = HWinfo()
        self.devices = []

        try:
            keys = self.hwinfo.keyboard['keylist']
        except:
            keys = []

        self.active_modifiers = set()
        self.keys = []
        for k in keys:
            key = Key()

            if 'evkey' in k:
                key.scancode = getattr(evdev.ecodes, 'KEY_' + k['evkey'])
                key.codes = [k['evkey']]
                if k['evkey'] in default_descriptions:
                    key.description = default_descriptions[k['evkey']]
            if 'location' in k:
                key.location = k['location']
            if 'modifiers' in k:
                key.modifiers = set(k['modifiers'])
            if 'devname' in k:
                key.devname = k['devname']
            if 'description' in k:
                key.description = k['description']

            if key.scancode != -1 and (key.modifiers or self.is_tested(key.scancode)):
                key.tested = True

            if 'tested' in k:
                key.tested = k['tested']

            self.keys.append(key)
            self.app_ctrl.send(KeySignal(key))

    def tearDown(self):
        for device in self.devices:
            try:
                device.ungrab()
            except:
                pass

        try:
            out = HWinfo(load=False)

            keys = []
            for k in self.keys:
                key = {}
                if k.codes:
                    key['evkey'] = k.codes[0]
                if k.location:
                    key['location'] = k.location
                if k.description:
                    key['description'] = k.description
                if k.devname:
                    key['devname'] = k.devname
                if k.modifiers:
                    key['modifiers'] = list(k.modifiers)

                if k.found:
                    keys.append(key)

            out.keyboard = { 'keylist' : keys }
            path = os.path.join(self.outputdir, 'evdev_keys.yaml')
            out.dump(open(path, 'w'))
        except:
            details = sys.exc_info()[1]
            self.log.error('Failed to write key list ({:s})!'.format(details))

        SessionTest.tearDown(self)


    def is_tested(self, scancode):
        # See kernel input-event-codes.h for more information
        # Err on the side of including a few too meny keys.
        # NOTE: Does not include currently unassigned keycodes!

        if evdev.ecodes.KEY_MUTE <= scancode <= evdev.ecodes.KEY_POWER:
            return True
        if evdev.ecodes.KEY_SCALE <= scancode <= evdev.ecodes.KEY_SCALE:
            return True
        if evdev.ecodes.KEY_STOP <= scancode <= evdev.ecodes.KEY_EDIT:
            return True
        if evdev.ecodes.KEY_NEW <= scancode <= evdev.ecodes.KEY_REDO:
            return True
        # F13 - F24
        if 183 <= scancode <= 194:
            return True
        if evdev.ecodes.KEY_PLAYCD <= scancode <= evdev.ecodes.KEY_MICMUTE:
            return True
        if evdev.ecodes.KEY_OK <= scancode <= evdev.ecodes.KEY_LOGOFF:
            return True
        if evdev.ecodes.KEY_FRAMEBACK <= scancode:
            # Note, includes:
            #  - Diagonal movement keys
            #  - Dollar/Euro keys
            #  - Numeric pads (phone like)
            #  - Generic Fn keys
            #  - ...
            return True

        return False

    def find_key_in_db(self, scancode, device, modifiers):
        for key in self.keys:
            if key.devname and key.devname != device:
                continue

            if set(key.modifiers) != modifiers:
                continue

            if key.scancode == scancode:
                return key

        return None

    def process_event(self, device, event):
        event = evdev.categorize(event)

        if not isinstance(event, evdev.KeyEvent):
            return

        # NOTE: Not quite clean, but in none of these cases we get a list
        if isinstance(event.keycode, str):
            event.keycode = [event.keycode, ]

        if event.scancode in modifier_keycodes:
            # Keep track of modifiers
            if event.keystate in [event.key_down, event.key_hold]:
                self.active_modifiers.add(event.scancode)
            if event.keystate in [event.key_up]:
                if event.scancode in self.active_modifiers:
                    self.active_modifiers.remove(event.scancode)
                else:
                    self.log.info('Ignoring modifier inactivation from {:s}'.format(str(event)))
            # Ignore any modifier key press
            return

        modifiers = set()
        for mod in self.active_modifiers:
            modifiers.add(modifier_keycodes[mod])

        non_modifier_key = self.is_tested(event.scancode)
        if non_modifier_key:
            modifiers = set()

        if non_modifier_key or \
           (modifiers and self.grabbed) or\
           self.find_key_in_db(event.scancode, device.name, modifiers):

            self.log.info("Got %s", str(event))

            codes = [c[4:] if c.startswith('KEY_') else c for c in event.keycode]

            found = True
            devname = device.name

            key = self.find_key_in_db(event.scancode, device.name, modifiers)
            if key is None:
                key = Key()
                key.scancode = event.scancode
                key.modifiers = modifiers
                key.dbmissing = True
                key.tested = True
                key.location = 'please enter'
                if codes[0] in default_descriptions:
                    key.description = default_descriptions[codes[0]]
                else:
                    key.description = ''
                self.keys.append(key)

            key.codes = codes
            key.devname = device.name
            key.found = True

            self.app_ctrl.send(KeySignal(key))
        else:
            self.log.info("Ignoring %s", str(event))

        return event

    @interactive
    def test(self):
        """
        Test plugging in with slow edid a number of times.
        """

        # If anything goes wrong in this test then most likely the input is
        # grabbed. In that case the user has no chance of escaping, so we
        # should just end the test in an ERROR situation. In all other cases
        # the user will have the chance to comment, etc. as part of the test.
        self.skip_user_confirm = True

        if not self.keys:
            self.log.info('No key information in database!')
            have_db = False
        else:
            have_db = True

        # Get a grab for all devices
        self.devices = []
        for device in evdev.list_devices():
            try:
                device = evdev.InputDevice(device)
            except OSError:
                # This happens e.g. on the X1 Yoga on RHEL 7.4 with the accelerometer
                self.log.warn("Could not open device %s" % device)
                continue

            self.devices.append(device)
            self.log.debug("Grabbing all input from device %s" % device.name)
            device.grab()

        self.app_ctrl.status.mark_action('grabbed')
        self.grabbed = True

        fds = {device.fileno() : device for device in self.devices }

        fds[self.app_ctrl.fileno()] = self.app_ctrl

        done = False
        while not done:
            # We don't have an async handler to find new devices, so probe every
            # 0.5 seconds.
            r, w, x = select.select(fds.keys(), [], fds.keys(), 0.5)
            for fd in r + x:
                obj = fds[fd]

                if obj is self.app_ctrl:
                    event = self.app_ctrl.get_event()

                    if self.app_ctrl.status.continued:
                        if not self.grabbed:
                            for device in self.devices:
                                device.grab()
                            self.grabbed = True
                        self.app_ctrl.status.mark_action('grabbed')
                        self.app_ctrl.status.allow_finish(allow=False)

                    if isinstance(event, KeyChangedEvent):
                        if event.scancode > 0:
                            key = self.find_key_in_db(event.scancode, event.devname, event.modifiers)
                            if key:
                                for k, v in event.changes.iteritems():
                                    setattr(key, k, v)
                                self.app_ctrl.send(KeySignal(key))

                    elif isinstance(event, KeyDeleteEvent):
                        if event.scancode > 0:
                            # It will already be deleted on the client, so we
                            # don't send a notification. Note that it is a fatal
                            # error if the key does not exist.
                            key = self.find_key_in_db(event.scancode, event.devname, event.modifiers)
                            self.keys.remove(key)

                else:
                    try:
                        event = fds[fd].read_one()
                    except IOError:
                        self.log.info('Device %s disappeared' % fds[fd].name)
                        self.devices.remove(fds[fd])
                        del fds[fd]

                        continue

                    evkey = self.process_event(fds[fd], event)

                    if not evkey or not self.grabbed:
                        # Nothing interesting
                        pass

                    elif 'KEY_ENTER' in evkey.keycode or \
                         'KEY_KPENTER' in evkey.keycode or \
                         'KEY_ESC' in evkey.keycode:
                        # Go into ungrabbed mode
                        for device in self.devices:
                            device.ungrab()

                        self.grabbed = False
                        self.app_ctrl.status.mark_action('query')
                        self.app_ctrl.status.allow_finish(allow=True)
                        self.app_ctrl.status.allow_continue()

            for device in evdev.list_devices():
                for d in fds.itervalues():
                    if isinstance(d, evdev.InputDevice) and d.fn == device:
                        break
                else:
                    try:
                        device = evdev.InputDevice(device)
                    except OSError:
                        # We probably warned already above, so ignore this here
                        continue
                    self.log.info("New device %s detected" % device.name)
                    self.devices.append(device)
                    fds[device.fileno()] = device
                    if self.grabbed:
                        self.log.debug("Grabbing input for new device")
                        device.grab()

            # Update the UI status
            missing = False
            for key in self.keys:
                if key.tested and not key.found:
                    missing = True

            if missing:
                self.set_ui_status('FAIL')
            elif have_db:
                self.set_ui_status('PASS')
            else:
                self.set_ui_status(None)

            # Check if user stopped the test
            if self.app_ctrl.status.finished:
                done = True

        missing = 0
        tested = 0
        count = 0
        ok = 0
        extra = 0
        seen = 0

        for key in self.keys:
            count += 1
            if key.dbmissing:
                if have_db:
                    self.log.warning("Key %s on %s is missing in DB" % (','.join(key.codes), key.devname))
                else:
                    self.log.info("Key %s on %s is missing in DB" % (','.join(key.codes), key.devname))
                missing += 1
            if key.tested:
                tested += 1
            if key.found and key.tested:
                ok += 1
            if not key.found and key.tested:
                self.log.error("Key %s was not pressed" % ','.join(key.codes))
            if key.found:
                seen += 1
            else:
                extra += 1

        self.whiteboard = 'Saw %i unique keys, %i of %i expected keys were found %i of these are missing in the DB.' % (seen, ok, tested, missing)

        if ok != tested:
            self.fail('Only found %i of %i keys!' % (ok, tested))

if __name__ == "__main__":
    main()

