#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

from __future__ import division

import os
import os.path
import re
import select
import numpy as np

from avocado import main
from avocado.core import exceptions

from fed_laptoptest.test import SessionTest, interactive
from fed_laptoptest.hwinfo import HWinfo
from fed_laptoptest.ui import Signal, Event, TestUI
from fed_laptoptest.utils import audio

from gi.repository import Gtk, GLib

class StartTestSignal(Signal):
    pass

class ResultEvent(Event):
    def __init__(self, result, whiteboard=None):
        self.result = result
        self.whiteboard = whiteboard

class SimpleAudioUI(TestUI):

    # Don't destroy ears
    MAX_OUTPUT_VOLUME = 0.6
    # Don't pick up high frequency noise as distortions
    HARMONICS = 9

    def set_test_params(self, signal_volume, output_gain, input_gain):
        self.test.set_volume(signal_volume)
        audio.set_output_volume(output_gain)
        audio.set_input_volume(input_gain)

    def done_cb(self, obj):
        gain = obj.get_gain()
        thd_f = obj.get_THD_F(self.HARMONICS + 1)

        if gain > 0:
            system_gain_db = (
                  20 * np.log10(gain)
                - 20 * np.log10(self.pending_tests[0][1]**3)
                - 20 * np.log10(self.pending_tests[0][2]**3))
        else:
            system_gain_db = -np.inf

        self.app.info('System gain: %6.1fdB (%4.3f, harmonics: %5.1f%%, output: %f, input: %f, wave: %f)' %
                      (system_gain_db, gain, 100*thd_f, self.pending_tests[0][1], self.pending_tests[0][2], self.pending_tests[0][0]))

        self.results.append((self.pending_tests.pop(0), gain, system_gain_db, thd_f))

        self.next_test()

    def find_gain(self):
        mid_gain = self.results[0][1]

        # Really low gain, give up.
        if mid_gain < 0.001:
            self.app.warning('Microphone does not seem to have picked up our test sound!')
            self.app.send(ResultEvent(False, 'Microphone did not pick up the test sound!'))
            return

        def testsrc_volume(mic, output):
            expected_gain = mid_gain / (0.5 ** 3 * 0.5 ** 3) * mic ** 3 * output ** 3

            if expected_gain > 0.8:
                target_gain = 0.8
            else:
                target_gain = expected_gain

            return min(1.0, 0.8 / expected_gain, self.MAX_OUTPUT_VOLUME**3 / output**3)

        # This is at 0.5, 0.5
        # Do two rounds trying to keep overall gain equal:
        #  1. Ramp output gain
        #  2. Ramp input gain
        # while limiting overall volume and gain to values that should not
        # horribly distort.
        for i in xrange(5):
            mic = 1.0
            output = 0.2 * (i + 1)
            volume = testsrc_volume(mic, output)
            self.pending_tests.append((volume, output, mic))

        for i in xrange(5):
            mic = 0.2 * (i + 1)
            output = 0.5
            volume = testsrc_volume(mic, output)
            self.pending_tests.append((volume, output, mic))

        self.results = []
        self._pending_finished = self.check_stable

        self.next_test()

    def check_stable(self):
        msgs = []

        # We want to allow a few outliers, but not too many
        system_gains_db = np.array([ t[2] for t in self.results ])
        thd_f = np.array([ t[3] for t in self.results ])

        system_gains_db = np.array(system_gains_db)

        if system_gains_db.std() > 6:
            self.app.warning('The systems input/output gain do not seem to be consistent (std=%f)!' % system_gains_db.std())
            msgs.append('Gain was not consistent!')
        else:
            self.app.info('The systems input/output gain is consistent (std=%f)!' % system_gains_db.std())

        if system_gains_db.mean() > 12:
            self.app.warning('The systems input/output gain seem to be very high (average: %fdB)! This likely means that the microphone gain is too high requiring kernel workarounds!' % system_gains_db.mean())
            msgs.append('Gain was extremely high!')
        else:
            self.app.info('The systems input/output gain averages at %fdB!' % system_gains_db.mean())

        # Only check distortions for second part of test where the output volume
        # is reasonably high (i.e. noise is lower relatively).
        thd_f = thd_f[5:]
        if np.max(thd_f) > 0.15:
            self.app.warning('The system appears to have very strong distortions (max(THD_F)=%f%%)!' % (100 * np.max(thd_f)))
            msgs.append('System has strong distortions!')
        else:
            self.app.info('Maximal THD_F: %f %% (first %d harmonics)!' % (100 * np.max(thd_f), self.HARMONICS))

        if msgs:
            whiteboard = ' '.join(msgs)
        else:
            whiteboard = None

        self.app.send(ResultEvent(max(system_gains_db) > -20, whiteboard))

    def next_test(self):
        if not self.pending_tests:
            self._pending_finished()
        else:
            self.set_test_params(*self.pending_tests[0])
            self.test.play()

    def recv(self, signal):
        if not isinstance(signal, StartTestSignal):
            return

        self.test = audio.SineAmplitudeTest(samples=40000)
        self.test.connect('done', self.done_cb)
        self._system_gains_db = []

        audio.set_input_mute(False)
        audio.set_output_mute(False)

        # Do a few initial measurements of the receiving amplitude
        self._pending_finished = self.find_gain
        self.results = []
        self.pending_tests = [
            ( 1.0, 0.5, 0.5 ),
        ]

        self.next_test()



class SimpleAudio(SessionTest):
    """
    .. title:: Simple audio speaker and microphone test

    This test will output some audio and tests whether it is received changing
    both the input and output volume.

    The test assume that the default sink are the devices speakers and the
    default source is an integrated microphone.

    1. .. class:: action-waiting

           Please ensure the following and press *Continue* to start the test:

           * The noise level is low (e.g. turn off music, close windows, pause conversations)
           * No headphones or external speakers are attached (including HDMI monitors)
           * The internal speakers and microphone are selected for playback and recording

           The default settings should be correct if nothing has been attached.

    2. .. class:: action-running

       Test is running, this will take only a few seconds.

    3. .. class:: action-finish

       Test has completed.

    :avocado: enable
    :avocado: tags=manual
    :categories: screen
    """

    ui_class = SimpleAudioUI

    @interactive
    def test(self):
        self.app_ctrl.status.mark_action('waiting')
        self.app_ctrl.status.wait_continue()

        self.app_ctrl.status.mark_action('running')
        self.app_ctrl.send(StartTestSignal())

        while True:
            event = self.app_ctrl.get_event()
            if isinstance(event, ResultEvent):
                if event.whiteboard:
                    self.whiteboard = event.whiteboard

                if event.result:
                    return
                else:
                    self.fail('Test Failed!')

if __name__ == "__main__":
    main()

