#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import os
import os.path
import re
import evdev
import select
import time

from avocado import main
from avocado.core import exceptions

from fed_laptoptest.test import SessionTest, interactive
from fed_laptoptest.hwinfo import HWinfo
from fed_laptoptest.ui import Signal, Event, TestUI
from fed_laptoptest.utils import network, ui as uiutils

from gi.repository import Gtk, GLib, NM, Gio

typemap = {
    NM.DeviceType.WIFI      : 'Wireless LAN',
    NM.DeviceType.ETHERNET  : 'Ethernet (Wired)',
}


class DevInfoSignal(Signal):
    def __init__(self, nmdev):
        # Note: On old version of NM (1.4.0) the enumerations cannot be pickled!

        props = ['path', 'interface', 'ip-interface', 'driver',
                 'driver-version', 'product', 'physical-port-id']
        if nmdev.get_device_type() == NM.DeviceType.ETHERNET:
            props += ['hw-address', 'perm-hw-address', 'carrier']
        elif nmdev.get_device_type() == NM.DeviceType.WIFI:
            props += ['hw-address', 'perm-hw-address', 'capabilities']

        self.nmdev = {}
        for prop in props:
            self.nmdev[prop] = nmdev.get_property(prop)

        self.nmdev['type'] = int(nmdev.get_device_type())

        if nmdev.get_device_type() == NM.DeviceType.WIFI:
            self.nmdev['capabilities'] = int(self.nmdev['capabilities'])
            active_connection = nmdev.get_active_connection()
            if active_connection is not None:
                state = active_connection.get_state()
            else:
                state = NM.ActiveConnectionState.DEACTIVATED
            self.nmdev['state'] = int(state)

        self.ping_info = nmdev.ping_info

class SimpleNetworkTestUI(TestUI):

    def recv(self, signal):
        if not isinstance(signal, DevInfoSignal):
            return

        # Find the interface, if it does not exist then add it
        for row in self.listbox.get_children():
            if row.path == signal.nmdev['path']:
                break
        else:
            row = Gtk.ListBoxRow()
            row.path = signal.nmdev['path']
            self.listbox.add(row)

        self.update_dev_info(signal, row)

    def update_dev_info(self, signal, row):
        nmdev = signal.nmdev
        hbox = row.get_child()
        if not hbox:
            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

            hbox.img = Gtk.Image()
            hbox.pack_start(hbox.img, fill=True, expand=False, padding=5)

            hbox.label = Gtk.Label(xalign=0)
            hbox.pack_start(hbox.label, fill=True, expand=True, padding=5)
            row.add(hbox)

        if nmdev['type'] == NM.DeviceType.ETHERNET:
            state = 'cable attached' if nmdev['carrier'] else 'no cable attached'
        elif nmdev['type'] == NM.DeviceType.WIFI:
            if nmdev['state'] == NM.ActiveConnectionState.ACTIVATING:
                state = 'activating'
            elif nmdev['state'] == NM.ActiveConnectionState.ACTIVATED:
                state = 'active'
            elif nmdev['state'] == NM.ActiveConnectionState.DEACTIVATING:
                state = 'deactivating'
            elif nmdev['state'] == NM.ActiveConnectionState.DEACTIVATED:
                state = 'deactive'
            else:
                state = 'unknown'

        hbox.img.set_from_gicon(uiutils.status_to_icon('PASS' if signal.ping_info else 'FAIL'), Gtk.IconSize.LARGE_TOOLBAR)

        hbox.label.set_markup(
            '<b>{type:s} Adapter {name:s} ({driver:s})</b>\n'
            'MAC: {mac:s}, driver version: {version:s}\n'
            'current status: {state:s}'.format(
                type=typemap[nmdev['type']],
                name=nmdev['product'],
                driver=nmdev['driver'],
                version=nmdev['driver-version'],
                mac=nmdev['perm-hw-address'],
                state=state
            ))

        row.show_all()

    def get_widget(self):
        scroll = Gtk.ScrolledWindow()
        self.listbox = Gtk.ListBox()

        scroll.add(self.listbox)
        scroll.show_all()

        return scroll



class SimpleNetworkTest(SessionTest):
    """
    Base class for simple network tests.

    The documentation of the derived classes should contain the following
    actions:
     * wait-user
     * finish

    :avocado: disable
    """

    ui_class = SimpleNetworkTestUI

    ADDRESS = '8.8.8.8'
    TYPES = [NM.DeviceType.UNKNOWN,]

    def runtest(self):
        context = GLib.main_context_default()

        self.app_ctrl.status.mark_action('wait-user')

        devices = network.nm_get_devices(self.TYPES)
        test_queue = []

        for device in devices:
            device.ping_info = False
            self.app_ctrl.send(DevInfoSignal(device))

        # If we didn't find any devices, skip to the next step
        if devices:
            while True:
                if self.app_ctrl.status.finished:
                    break

                # Refresh device information on client
                for d in devices:
                    self.app_ctrl.send(DevInfoSignal(d))

                while context.iteration(False):
                    # Sometimes old NM versions unref too often, so just reference again.
                    # Note that the the check is for 1.4.4 as this version is known to
                    # work. 1.4.0 is known to not work and the fix might have happened
                    # before 1.4.4.
                    if NM.MAJOR_VERSION == 1 and NM.MINOR_VERSION == 4 and NM.MICRO_VERSION < 4:
                        context.ref()
                self.app_ctrl.read_pending()

                if not test_queue:
                    test_queue = [d for d in devices if d.ping_info is False]

                if not test_queue:
                    # All devices tested successfully
                    self.app_ctrl.status.mark_action('finish')
                    self.app_ctrl.status.allow_finish()
                    time.sleep(0.5)
                    continue
                else:
                    self.app_ctrl.status.mark_action('wait-user')

                # Test one device at a time (just to not block indefiniately)
                device = test_queue.pop()

                interface = device.props.ip_interface
                if not interface:
                    interface = device.props.interface
                if interface:
                    if network.ping(self.ADDRESS, interface):
                        device.ping_info = True
                        self.app_ctrl.send(DevInfoSignal(device))

                # Throttle things a bit
                time.sleep(0.2)


class SimpleEthernetTest(SimpleNetworkTest):
    """
    .. title:: Basic Ethernet Functionality

    This test tries to do a simple network test on all detected ethernet
    adapters. The detected ethernet ports and their state will be shown. If
    internet connectivity (using a simple ping) is detected, then it will be
    marked as working.

    This test currently pings the IPv4 address 8.8.8.8.

    Please make sure that every port you have was detected. If a port is not
    detected then mark it as expected.

    1. .. class:: action-wait-user

       Running detection test on all ports. Please connect each port physically
       and configure an internet connection. You can use the GNOME Control
       Center to do so.

    2. .. class:: action-finish

        Network was detected on all ports. Please confirm that there are no
        further ports.

    :avocado: enable
    :avocado: tags=manual
    :categories: ethernet,networking
    """

    TYPES = [NM.DeviceType.ETHERNET,]

    # Need an entry point in the subclass
    @interactive
    def test(self):
        self.runtest()


class SimpleWIFITest(SimpleNetworkTest):
    """
    .. title:: Basic WIFI Functionality

    This test tries to do a simple network test on all detected wireless
    adapters. The detected wireless interfaces are shown and if internet
    connectivity (using a simple ping) is detected, then it will be
    marked as working.

    This test currently pings the IPv4 address 8.8.8.8.

    Please make sure that every port you have was detected. If a port is not
    detected and 

    1. .. class:: action-wait-user

       Running detection test on all wireless network adapters. If you have
       a wireless network available then please configure it now using either
       GNOME Control Center or the Network Manager connection editor.

    2. .. class:: action-finish

        Internet connectivity was detected on all wifi devices. Please confirm
        there are no further undetected devices.

    :avocado: enable
    :avocado: tags=manual
    :categories: wireless,networking
    """

    TYPES = [NM.DeviceType.WIFI,]

    # Need an entry point in the subclass
    @interactive
    def test(self):
        self.runtest()


if __name__ == "__main__":
    main()

